#include "PointingIK.h"

#include <VirtualRobot/Robot.h>
#include <VirtualRobot/RobotNodeSet.h>

#include <ArmarXCore/core/Component.h>

namespace armarx::control::pointing::core
{

    PointingIK::PointingIK(VirtualRobot::RobotPtr robot, Side side)
    {
        auto humanMapping = robot->getHumanMapping();
        ARMARX_CHECK(humanMapping.has_value())
            << "Robot '" << robot->getName() << "' has no human mapping";

        auto armDescription = (side == LEFT) ? humanMapping->leftArm : humanMapping->rightArm;

        auto elbow = armDescription.segments.elbow.nodeName;
        auto wrist = armDescription.segments.wrist.nodeName;

        auto nodeSetArm = robot->getRobotNodeSet(armDescription.nodeSet);
        auto joints = nodeSetArm->getNodeNames();
        joints.resize(nodeSetArm->getRobotNodeIndex(wrist));


        reducedRobot_ = std::make_unique<simox::control::simox::robot::Robot>(
            robot,
            joints,
            std::vector(
                {wrist})); // TODO true flag for reduced -> fueht zu self coll? (auch ohne ...)

        ik_ = std::make_unique<simox::control::method::example::PointingIK>(
            *reducedRobot_,
            *reducedRobot_->getNode(
                wrist), // TODO TCP instead of wrist? IK parameter is named "wrist"...
            *reducedRobot_->getNode(elbow),
            side == LEFT ? reducedRobot_->humanMapping()->leftArm
                         : reducedRobot_->humanMapping()->rightArm);
    }

    TrajectoryPtr
    PointingIK::computeTrajectory(Eigen::Vector3f& target) // self coll
    {
        ik_->setPointingTarget(target.cast<double>() / 1000); // simox uses meters

        // #joints x #timestamps
        std::vector<std::vector<double>> data(ik_->getNodeNames().size());
        auto appendPositions = [&data](Eigen::VectorXd pos)
        {
            for (size_t i = 0; i < data.size(); i++)
            {
                data[i].push_back(pos[i]);
            }
        };
        appendPositions(ik_->getJointValues());

        Eigen::VectorXd prevJointValues = ik_->getJointValues();
        while (not ik_->isTargetReached())
        {
            ik_->optimize();

            if ((ik_->getJointValues() - prevJointValues).norm() < 1e-6)
            {
                return nullptr;
            }
            prevJointValues = ik_->getJointValues();

            appendPositions(ik_->getJointValues());
        }
        return new Trajectory(data, Ice::DoubleSeq(), ik_->getNodeNames()); // TODO memory leak?
    }

} // namespace armarx::control::pointing::core
