/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    control::ArmarXObjects::retrieve_hand
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <RobotAPI/libraries/armem_objects/client/attachment/Reader.h>
#include <RobotAPI/libraries/armem_objects/client/attachment/Writer.h>
#include <RobotAPI/libraries/armem_robot_state/client/common/VirtualRobotReader.h>
#include <RobotAPI/libraries/robot_name_service/core/Robot.h>


namespace armarx::control::retrieve_hand::core
{
    /**
    * @defgroup Library-retrieve_hand retrieve_hand
    * @ingroup control
    * A description of the library retrieve_hand.
    *
    * @class retrieve_hand
    * @ingroup Library-retrieve_hand
    * @brief Brief description of class retrieve_hand.
    *
    * Detailed description of class retrieve_hand.
    */
    class RetrieveHand
    {
    public:

        struct Remote
        {
            // armarx::viz::Client arviz;
            armarx::armem::client::MemoryNameSystem memoryNameSystem;

            armarx::RobotUnitInterfacePrx robotUnit;
        };

        struct Properties
        {
            std::string robotName;
            std::string robotNodeSet;
            double timeDurationInSec;
            // Eigen::Matrix4f targetPoseInRoot;
            armarx::FramedPose targetPose;
            float kpLinear = 10.0f;
            float kpAngular = 10.0f;
            float kdLinear = 1.0f;
            float kdAngular = 1.0f;

            std::vector<std::string> fileNames;
        };

        struct Subskills
        {
            std::function<bool(const Eigen::Vector3f&)> lookAt;
            std::function<bool()> openHand;
            std::function<bool()> closeHand;
        };

        RetrieveHand(const Remote&, const Properties&);
        virtual ~RetrieveHand() = default;

        void
        stop()
        {
            running.store(false);
            _deleteTSVMPController();
        }

        bool execute()
        {
            std::scoped_lock l(executionMutex);
            running.store(true);
            return _runTSVMPController();
        }

    protected:
        Remote remote;
        Properties properties;

    private:
        bool _runTSVMPController();
        void _deleteTSVMPController();
        mutable std::mutex executionMutex;
        std::atomic_bool running = false;
    };

}
